/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */


#ifndef COMMON_H
#define COMMON_H  1

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include <alsa/asoundlib.h>
#include <math.h>

/* Number of RTTY data elements (bits) per character */
#define DATABITS_PER_CHAR	5

/* Number of signal level samples per RTTY element (bit) */
#define LEVELS_PER_BIT		4

/* Number of RTTY characters buffered in decoder */
#define NUM_BUFFERED_CHARS	3

/* Size of char arrays (strings) for error messages etc */
#define MESG_SIZE	128

/* Control flags */
#define TRANSMIT_TAG		0x00000001 /* Send tag string  */
#define ENABLE_CAT			0x00000002 /* Enable Trcvr CAT */
#define CAT_SETUP			0x00000004 /* Trcvr CAT setup OK  */
#define ENABLE_RTS_DTR		0x00000008 /* Enable RTS control of PTT */
#define ENABLE_DTR			0x00000010 /* Enable DTR control of PTT */
#define RTS_DTR_SETUP		0x00000020 /* RTS/DTR control setup OK */
#define RECORD_QSO			0x00000040 /* Enable recording of QSO's   */
#define RECEIVE_MODE		0x00000080 /* xritty is in Receive mode   */
#define TRANSMIT_MODE		0x00000100 /* xritty is in Transmit mode  */
#define TRANSMIT_MACRO		0x00000200 /* Start transmission of Macro */
#define TRANSMIT_KEYBD		0x00000400 /* Transmission from Keyboard  */
#define ENABLE_USOS			0x00000800 /* Enable USOS (Un Shift On Space) */
#define SQUELCH_OPEN		0x00001000 /* Signal level allows squelch to open */
#define CAPTURE_SETUP		0x00002000 /* Sound Capture has been set-up  */
#define PLAYBACK_SETUP		0x00004000 /* Sound Playback has been set-up */
#define MIXER_SETUP			0x00008000 /* Sound Mixer has been set-up */
#define SAVE_RECORD			0x00010000 /* Record was saved to file    */
#define ENABLE_DISPLAY		0x00020000 /* Enable display of sync train status */
#define TRANSMIT_MORSE_ID	0x00040000 /* Transmit callsighn in Morse code */
#define TCVR_SERIAL_TEST	0x00080000 /* Transceiver serial port under test */
#define LTRS_SHIFT			0x00100000 /* Transmitting in LTRS shift mode */
#define REVERSE_AFSK		0x00200000 /* Use Revese AFSK frequency shift */
#define ENABLE_DIVERSITY	0x00400000 /* Enable sudo-diversity in RTTY decoder */

/* Some character definitions */
#define HT				0x09 /* Horizontal Tab  */
#define CR				0x0d /* Carriage Return */
#define LF				0x0a /* Line Feed       */
#define NO_CHARACTER	0x80 /* No rtty character decoded yet */

/* Return values */
#define SUCCESS		0
#define ERROR		1

/* For the Error Dialog */
#define QUIT	TRUE
#define OK		FALSE

/* 'NUL' character used as 'last character' flag */
#define NULL_CHAR	0x00

/* Number and size of labels for commands  */
#define NUM_OF_LABELS	10
#define LABEL_LENGTH	24

/* Should have been in math.h */
#ifndef M_2PI
  #define M_2PI		6.28318530717958647692
#endif

/*
 * Standard gettext macros.
 */
#ifdef ENABLE_NLS
#  include <libintl.h>
#  undef _
#  define _(String) dgettext (PACKAGE, String)
#  define Q_(String) g_strip_context ((String), gettext (String))
#  ifdef gettext_noop
#    define N_(String) gettext_noop (String)
#  else
#    define N_(String) (String)
#  endif
#else
#  define textdomain(String) (String)
#  define gettext(String) (String)
#  define dgettext(Domain,Message) (Message)
#  define dcgettext(Domain,Message,Type) (Message)
#  define bindtextdomain(Domain,Directory) (Domain)
#  define _(String) (String)
#  define Q_(String) g_strip_context ((String), (String))
#  define N_(String) (String)
#endif

/*-------------------------------------------------------------------*/

/* Configuration parameters structure */
typedef struct
{
  char
	snd_card[32], /* Sound card name */
	pcm_dev[32],  /* PCM device name */
	cap_src[32],  /* Capture source  */
	cap_vol[32],  /* Capture volume control  */
	mst_vol[32],  /* Master volume control   */
	pcm_vol[32];  /* Playback volume control */

  int
	channel,	/* ALSA "channel" for use in mixer setup */
	num_chn,	/* Number of audio channels (2=stereo, 1=mono) */
	use_chn,	/* Channel in use (frontleft=0, frontright=1 etc) */
	cap_lev,	/* Recording/Capture level */
	pbk_lev,	/* Playback (volume) level */
	tx_output,	/* Tx output power level */
	dsp_rate;	/* DSP rate (speed) samples/sec */

  /* Transmit buffer size */
  size_t xmit_buff_size;

  /* Transceiver type 1=FT847 2=FT857 3=K2 4=K3 0=NONE */
  int tcvr_type;

  char
	cat_serial[32], /* CAT serial port */
	rts_serial[32]; /* RTS serial port */

  char font[32]; /* Font to use in text views */

  int
	tone_dur,			/* Preamble tone duration in secs */
	samples_per_char,	/* Total signal samples per RTTY character */
	rtty_bit,			/* Length of an RTTY element (bit) in DSP samples */
	rtty_2bit,			/* Length * 2 of an RTTY element (bit) in DSP samples */
	rtty_bit4,			/* Length/4 of an RTTY element (bit) in DSP samples */
	stop_bit_len,		/* Length of stop bit in DSP samples */
	afsk_shift,			/* Mark/Space frequency shift in Hz */
	mark_freq,			/* AFSK mark frequency */
	space_freq,			/* AFSK space frequency */
	tone_freq,			/* Selected Receiver BFO tone frequency */
	sqlch_thr,			/* Squelch threshold */
	morse_wpm,			/* Morse encoder speed: wpm */
	word_wrap,			/* Word wrap column number  */
	ifft_stride;		/* IFFT stride over input data (dsp samples) */

  double
	stop_bit,	/* Stop Bit duration (1.0, 1.5 or 2.0 bits) */
	mark_dW,	/* Phase increment for mark signal phase angle */
	space_dW,	/* Phase increment for space signal phase angle */
	tone_dW,	/* Phase increment for tone signal phase angle */
	baud_rate;	/* Baud Rate for transmission and reception */

  /* Local operator data */
  char
	call[15],
	name[13],
	zone[15],
	qth[13],
	loc[7];

  /* Tx/Rx power, N.Fig, antennas */
  char
	tx[12],
	tx_power[10],
	tx_ant[16],
	rx[12],
	rx_nfig[8],
	rx_ant[16];

	/* Xritty builder's glade file */
  char xritty_glade[128];

} rc_data_t;

/* QSO record type definition */
typedef struct
{
  /* QSO data */
  char
	dx_call[15],  /* Remote (dx) station callsign */
	dx_rst[4],    /* My RST report to Dx station  */
	my_rst[4],    /* My RST report to Dx station  */
	dx_name[13],  /* Remote station operator name */
	dx_qth[13],   /* Remote station QTH loc. name */
	dx_loc[7],    /* Remote station QTH locator   */
	date[12],     /* Normal format QSO date  */
	date_adif[9], /* QSO date in ADIF format */
	time[6],      /* Normal format QSO time  */
	time_adif[5], /* QSO time in ADIF format */
	freq[14],     /* QSO operating frequency MHz */
	mode[12],     /* Mode of operation */
	via[13];      /* Via which path or relay Sat */

  /* Remarks on QSO */
  char remarks[46];

} qso_record_t;

/*-------------------------------------------------------------------*/

/* Transceiver status data */
typedef struct
{
  int
	comp_lev,	/* Compressor Level */
	tx_power,	/* Tx Power Output  */
	vfo_freq,	/* VFO A frequency  */
	rig_mode;	/* Transceiver mode (usb|cw etc) */

} tcvr_status_t;

/*-------------------------------------------------------------------*/

/* Function prototypes produced by cproto */
/* callbacks.c */
void Error_Dialog(char *mesg, gboolean hide);
void Quit_Dialog(char *mesg);
void Save_Dialog(char *mesg);
void Bad_Entry_Dialog(char *mesg);
void on_main_window_destroy(GObject *object, gpointer user_data);
gboolean on_main_window_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
gboolean on_tx_textview_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
gboolean on_tx_textview_key_press_event(GtkWidget *widget, GdkEventKey *event, gpointer user_data);
void on_xmit_clicked(GtkButton *button, gpointer user_data);
void on_txvol_spinbutton_value_changed(GtkSpinButton *spinbutton, gpointer user_data);
gboolean on_rx_textview_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
void on_rcve_clicked(GtkButton *button, gpointer user_data);
void on_rxvol_spinbutton_value_changed(GtkSpinButton *spinbutton, gpointer user_data);
void on_squelch_value_changed(GtkRange *range, gpointer user_data);
gboolean on_scope1_draw(GtkWidget *widget, cairo_t *cr, gpointer user_data);
gboolean on_scope2_draw(GtkWidget *widget, cairo_t *cr, gpointer user_data);
gboolean on_scope1_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
void on_scope1_configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer user_data);
void on_scope2_configure_event(GtkWidget *widget, GdkEventConfigure *event, gpointer user_data);
void on_scope_label_button_press_event(GtkWidget *widget, GdkEventButton *event, gpointer user_data);
void on_rst_out_changed(GtkEditable *editable, gpointer user_data);
void on_rst_in_changed(GtkEditable *editable, gpointer user_data);
void on_op_name_changed(GtkEditable *editable, gpointer user_data);
void on_qth_name_changed(GtkEditable *editable, gpointer user_data);
void on_locator_changed(GtkEditable *editable, gpointer user_data);
void on_band_changed(GtkEditable *editable, gpointer user_data);
void on_callsign_changed(GtkEditable *editable, gpointer user_data);
void on_new_record_clicked(GtkButton *button, gpointer user_data);
void on_save_record_clicked(GtkButton *button, gpointer user_data);
void on_clear_fields_clicked(GtkButton *button, gpointer user_data);
void on_clear_windows_clicked(GtkButton *button, gpointer user_data);
void on_macro1_clicked(GtkButton *button, gpointer user_data);
void on_macro2_clicked(GtkButton *button, gpointer user_data);
void on_macro3_clicked(GtkButton *button, gpointer user_data);
void on_macro4_clicked(GtkButton *button, gpointer user_data);
void on_macro5_clicked(GtkButton *button, gpointer user_data);
void on_macro6_clicked(GtkButton *button, gpointer user_data);
void on_macro7_clicked(GtkButton *button, gpointer user_data);
void on_macro8_clicked(GtkButton *button, gpointer user_data);
void on_macro9_clicked(GtkButton *button, gpointer user_data);
void on_macro10_clicked(GtkButton *button, gpointer user_data);
void on_error_dialog_destroy(GObject *object, gpointer user_data);
gboolean on_error_dialog_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
void on_error_quit_button_clicked(GtkButton *button, gpointer user_data);
void on_error_ok_button_clicked(GtkButton *button, gpointer user_data);
void on_cancel_save_clicked(GtkButton *button, gpointer user_data);
void on_save_ok_clicked(GtkButton *button, gpointer user_data);
void on_bd45_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_bd50_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_bd75_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_bd100_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk85_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk170_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk200_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk250_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk450_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_afsk850_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_stop10_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_stop15_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_stop20_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_usos_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_k2_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_k3_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_ft847_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_ft857_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_rts_ctl_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_none_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_record_qsos_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_clear_window_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_identify_in_cw_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_quit_activate(GtkMenuItem *menuitem, gpointer user_data);
gboolean on_bad_entry_dialog_delete_event(GtkWidget *widget, GdkEvent *event, gpointer user_data);
void on_bad_entry_ok_clicked(GtkButton *button, gpointer user_data);
void on_cancel_quit_clicked(GtkButton *button, gpointer user_data);
void on_quit_ok_clicked(GtkButton *button, gpointer user_data);
void on_reverse_afsk_activate(GtkMenuItem *menuitem, gpointer user_data);
void on_shift_clicked(GtkButton *button, gpointer user_data);
void on_enable_diversity_activate(GtkMenuItem *menuitem, gpointer user_data);
/* cat.c */
gboolean Open_Tcvr_Serial(void);
void Close_Tcvr_Serial(void);
gboolean Open_RTS_Serial(void);
gboolean Read_Rx_Status(char *freq, char *rst);
gboolean Set_Tcvr_Mode(void);
gboolean Tune_Tcvr(GdkEventButton *event);
gboolean Set_Tx_Power(int tx_power, int comp_lev);
/* codec.c */
gboolean Decode_RTTY_Character(int *dec_char);
void Display_Scope1(cairo_t *cr);
void Display_Scope2(cairo_t *cr);
gboolean Transmit_Preamble(void);
gboolean Transmit_Postamble(void);
gboolean Transmit_RTTY_Character(int ascii_chr);
gboolean Morse_Transmit(char *mesg);
gboolean Make_Cosine_Table(void);
/* detect.c */
gboolean Afsk_Detect(unsigned char *mark_signal, unsigned char *space_signal);
/* display.c */
void Display_Waterfall(void);
void Display_Synchro(cairo_t *cr, int sync_pos, int sync_lev);
void Display_Signal(cairo_t *cr, GdkPoint *points);
void Set_TxRx_Labels(void);
void Clear_Record_Fields(gboolean all);
void Print_Character(int printchr, GtkTextBuffer *text_buffer, GtkWidget *scrolledwindow, GtkTextIter *iter);
void Set_Shift_Label(void);
/* ifft.c */
gboolean Initialize_IFFT(int16_t width);
void IFFT_Real(int16_t *data );
void IFFT_Data(int16_t sample);
/* interface.c */
GtkWidget *Builder_Get_Object(GtkBuilder *builder, gchar *name);
GtkWidget *create_main_window(GtkBuilder **builder);
GtkWidget *create_error_dialog(GtkBuilder **builder);
GtkWidget *create_save_dialog(GtkBuilder **builder);
GtkWidget *create_popup_menu(GtkBuilder **builder);
GtkWidget *create_bad_entry_dialog(GtkBuilder **builder);
GtkWidget *create_quit_dialog(GtkBuilder **builder);
/* main.c */
int main(int argc, char *argv[]);
/* operation.c */
gboolean Transmit_Keybd(gpointer data);
gboolean Transmit_Macro(gpointer data);
gboolean Receive_Mode(gpointer data);
gboolean Initialize_Transmit(gpointer data);
/* shared.c */
/* sound.c */
gboolean Open_Capture(char *mesg, int *error);
gboolean Open_Playback(char *mesg, int *error);
gboolean pen_Mixer(char *mesg, int *error);
gboolean Set_Playback_Level(int level, char *mesg, int *error);
gboolean Set_Capture_Level(int level, char *mesg, int *error);
void Close_Playback(void);
void Close_Capture(void);
void Close_Mixer(void);
gboolean Signal_Sample(short *sample_val);
gboolean DSP_Write(short *buffer, int buff_size);
/* utils.c */
int isFlagSet(int flag);
int isFlagClear(int flag);
void SetFlag(int flag);
void ClearFlag(int flag);
void ToggleFlag(int flag);
void Usage(void);
void Cleanup(void);
gboolean Load_Config(gpointer data);
gboolean Save_QSO_Record(void);
void Process_RST(GtkEditable *editable);
void Select_Macro(int macro_num);
gboolean Read_QSO_Record(void);
gboolean mem_alloc(void **ptr, size_t req);
gboolean mem_realloc(void **ptr, size_t req);
void free_ptr(void **ptr);
void Strlcpy(char *dest, const char *src, size_t n);
void Strlcat(char *dest, const char *src, size_t n);
double Strtod(const char *nptr);
void Init_Config(void);
gboolean Create_Wfall_Pixbuf(int width, int height);

#endif

