/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "detect.h"
#include "shared.h"

/*------------------------------------------------------------------------*/

/*  Afsk_Detect()
 *
 *  Detects the level of mark and space AFSK frequencies
 */

  gboolean
Afsk_Detect(
	unsigned char *mark_signal,
	unsigned char *space_signal )
{
  static int
	rtty_bit  = 0, /* For initialization */
	mark_freq = 0,
	buff_idx;	   /* Sample buffer index */

  int idx;

  /* Circular signal samples buffer for level detector */
  static short *samples_buff = NULL;

  /* Variables for the Goertzel algorithm */
  static double space_cosw, space_coeff;
  static double mark_cosw, mark_coeff;
  double space_q0, space_q1, space_q2;
  double mark_q0, mark_q1, mark_q2;
  double scale, sig;


  /* Initialize on change of parameters. Change
   * of mark implies change of space frequency too */
  if( !rc_data.mark_freq || !rc_data.rtty_bit )
	return( FALSE );
  if( (mark_freq != rc_data.mark_freq ) ||
	  (rtty_bit  != rc_data.rtty_bit ) )
  {
	mark_freq = rc_data.mark_freq;
	rtty_bit  = rc_data.rtty_bit;

	/* Omega for the mark frequency */
	double w;
	w = M_2PI * (double)rc_data.mark_freq /
	  (double)rc_data.dsp_rate;
	mark_cosw  = cos( w );
	mark_coeff = 2.0 * mark_cosw;

	/* Omega for the space frequency */
	w = M_2PI * (double)rc_data.space_freq /
	  (double)rc_data.dsp_rate;
	space_cosw  = cos( w );
	space_coeff = 2.0 * space_cosw;

	/* Allocate samples buffer for one RTTY bit, and clear */
	if( !mem_realloc((void **)&samples_buff,
		sizeof(short) * (size_t)rc_data.rtty_bit) )
	{
	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );

	  return( FALSE );
	}
	bzero( samples_buff,
		sizeof(short) * (size_t)rc_data.rtty_bit );
	buff_idx = 0;

  } /* if( (afsk_shift != rc_data.afsk_shift ) ||... ) */

  /* Save samples for a 1/4 of RTTY element (bit) */
  for( idx = 0; idx < rc_data.rtty_bit4; idx++ )
  {
	/* Get signal sample from buffer, abort on error */
	if( !Signal_Sample(&samples_buff[buff_idx]) )
	  return( FALSE );

	/* Increment/reset circular buffer's index */
	buff_idx++;
	if( buff_idx >= rc_data.rtty_bit )
	  buff_idx = 0;

  } /* for( idx = 0; idx < rc_data.rtty_bit4; idx++ ) */

  /* Calculate signal level of black and white
   * tone frequencies using Goertzel algorithm */
  space_q1 = space_q2 = 0.0;
  mark_q1  = mark_q2  = 0.0;
  for( idx = 0; idx < rc_data.rtty_bit; idx++ )
  {
	space_q0 =
	  space_coeff * space_q1 - space_q2 +
	  (double)samples_buff[buff_idx];
	space_q2 = space_q1;
	space_q1 = space_q0;

	mark_q0 =
	  mark_coeff * mark_q1 - mark_q2 +
	  (double)samples_buff[buff_idx];
	mark_q2 = mark_q1;
	mark_q1 = mark_q0;

	/* Increment/reset circular buffers' index */
	buff_idx++;
	if( buff_idx >= rc_data.rtty_bit )
	  buff_idx = 0;

  } /* for( idx = 0; idx < rc_data.rtty_bit; idx++ ) */

  /* Magnitude of space tone scaled by tone freq */
  scale = SIGNAL_SCALE * rc_data.rtty_bit;
  space_q1 /= scale;
  space_q2 /= scale;
  sig = space_q1 * space_q1 + space_q2 * space_q2 -
		space_q1 * space_q2 * space_coeff;
  if( sig > 255.0 ) sig = 255.0;
  *space_signal = (unsigned char)sig;

  /* Magnitude of mark tone scaled by tone freq */
  mark_q1 /= scale;
  mark_q2 /= scale;
  sig = mark_q1 * mark_q1 + mark_q2 * mark_q2 -
		mark_q1 * mark_q2 * mark_coeff;
  if( sig > 255.0 ) sig = 255.0;
  *mark_signal = (unsigned char)sig;

  return( TRUE );
} /* End of Afsk_Detect() */

/*------------------------------------------------------------------------*/

