/*
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "utils.h"
#include "shared.h"

/*------------------------------------------------------------------------*/

/*  Load_Line()
 *
 *  Loads a line from a file, aborts on failure. Lines beginning
 *  with a '#' are ignored as comments. At the end of file EOF is
 *  returned. Lines assumed maximum 80 characters long.
 */

  static int
Load_Line( char *buff, FILE *pfile, char *mesg )
{
  int
	num_chr, /* Number of characters read, excluding lf/cr */
	chr;     /* Character read by getc() */
  char error_mesg[MESG_SIZE];

  /* Prepare error message */
  snprintf( error_mesg, sizeof(error_mesg),
	  _("Error reading %s\n"\
		"Premature EOF (End Of File)"), mesg );

  /* Clear buffer at start */
  buff[0] = '\0';
  num_chr = 0;

  /* Get next character, return error if chr = EOF */
  if( (chr = fgetc(pfile)) == EOF )
  {
	fprintf( stderr, "xritty: %s\n", error_mesg );
	fclose( pfile );
	Error_Dialog( error_mesg, QUIT );
	return( EOF );
  }

  /* Ignore commented lines, white spaces and eol/cr */
  while(
	  (chr == '#') ||
	  (chr == ' ') ||
	  (chr == HT ) ||
	  (chr == CR ) ||
	  (chr == LF ) )
  {
	/* Go to the end of line (look for LF or CR) */
	while( (chr != CR) && (chr != LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xritty: %s\n", error_mesg );
		fclose( pfile );
		Error_Dialog( error_mesg, QUIT );
		return( EOF );
	  }

	/* Dump any CR/LF remaining */
	while( (chr == CR) || (chr == LF) )
	  /* Get next character, return error if chr = EOF */
	  if( (chr = fgetc(pfile)) == EOF )
	  {
		fprintf( stderr, "xritty: %s\n", error_mesg );
		fclose( pfile );
		Error_Dialog( error_mesg, QUIT );
		return( EOF );
	  }

  } /* End of while( (chr == '#') || ... */

  /* Continue reading characters from file till
   * number of characters = 80 or EOF or CR/LF */
  while( num_chr < 80 )
  {
	/* If LF/CR reached before filling buffer, return line */
	if( (chr == LF) || (chr == CR) ) break;

	/* Enter new character to line buffer */
	buff[num_chr++] = (char)chr;

	/* Get next character */
	if( (chr = fgetc(pfile)) == EOF )
	{
	  /* Terminate buffer as a string if chr = EOF */
	  buff[num_chr] = '\0';
	  return( SUCCESS );
	}

	/* Abort if end of line not reached at 80 char. */
	if( (num_chr == 80) && (chr != LF) && (chr != CR) )
	{
	  /* Terminate buffer as a string */
	  buff[num_chr] = '\0';
	  snprintf( error_mesg, sizeof(error_mesg),
		  _("Error reading %s\n"\
			"Line longer than 80 characters"), mesg );
	  fprintf( stderr, "xritty: %s\n%s\n", error_mesg, buff );
	  fclose( pfile );
	  Error_Dialog( error_mesg, QUIT );
	  return( ERROR );
	}

  } /* End of while( num_chr < max_chr ) */

  /* Terminate buffer as a string */
  buff[num_chr] = '\0';

  return( SUCCESS );
} /* End of Load_Line() */

/*------------------------------------------------------------------*/

/*  Load_Config()
 *
 *  Loads the xrittyrc configuration file
 */

  gboolean
Load_Config( gpointer data )
{
  int
	lb_idx, /* Label buffer index */
	mc_idx, /* Macro buffer index */
	ln_idx, /* Line buffer index  */
	ln_len, /* Line buffer length */
	tg_cnt, /* Count '<' and '>'  */
	idx;

  char
	rc_fpath[64], /* File path to xrittyrc */
	line[81];     /* Buffer for Load_Line  */

  /* Config file pointer */
  FILE *xrittyrc;

  /* Set text view font */
  PangoFontDescription *font_desc;

  /* For setting squelch level */
  GtkWidget *hscale;
  GtkAdjustment *adjustment;

  /* ALSA channel names and values */
  char *chan_name[] =
  {
	"FRONT_LEFT",
	"FRONT_RIGHT",
	"REAR_LEFT",
	"REAR_RIGHT",
	"SIDE_LEFT",
	"SIDE_RIGHT",
	"MONO"
  };

  int chan_number[] =
  {
	SND_MIXER_SCHN_FRONT_LEFT,
	SND_MIXER_SCHN_FRONT_RIGHT,
	SND_MIXER_SCHN_REAR_LEFT,
	SND_MIXER_SCHN_REAR_RIGHT,
	SND_MIXER_SCHN_SIDE_LEFT,
	SND_MIXER_SCHN_SIDE_RIGHT,
	SND_MIXER_SCHN_MONO
  };

  int num_chn = 7;

  /* Setup file path to xrittyrc */
  snprintf( rc_fpath, sizeof(rc_fpath),
	  "%s/xritty/xrittyrc", getenv("HOME") );

  /* Open xrittyrc file */
  xrittyrc = fopen( rc_fpath, "r" );
  if( xrittyrc == NULL )
  {
	perror( rc_fpath );
	Error_Dialog(
		_("Failed to open xrittyrc file\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  /*** Read Sound Card configuration data ***/
  /* Read card name, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Sound Card Name")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.snd_card, line, sizeof(rc_data.snd_card) );

  /* Read pcm device name, abort if EOF */
  if( Load_Line(line, xrittyrc, _("PCM Device Name")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.pcm_dev, line, sizeof(rc_data.pcm_dev) );

  /* Read DSP rate Samples/sec, abort if EOF */
  if( Load_Line(line, xrittyrc, _("DSP Rate") ) != SUCCESS )
	return( FALSE );
  rc_data.dsp_rate = atoi( line );

  /* Read ALSA "channel", abort if EOF */
  if( Load_Line(line, xrittyrc, _("ALSA Channel") ) != SUCCESS )
	return( FALSE );
  for( idx = 0; idx < num_chn; idx++ )
	if( strcmp(chan_name[idx], line) == 0 )
	  break;
  if( idx == num_chn )
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Invalid ALSA channel name\n"\
		  "Quit and correct xrittyrc"), QUIT);
	return( FALSE );
  }
  rc_data.channel = chan_number[idx];

  /* Set right or left channel buffer index */
  if( strstr(line, _("LEFT")) || strstr(line, _("MONO")) )
	rc_data.use_chn = 0;
  else
	rc_data.use_chn = 1;

  /* Set number of channels */
  if( strstr(line, _("MONO")) )
	rc_data.num_chn = 1;
  else
	rc_data.num_chn = 2;

  /* Read capture source, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Capture Source")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cap_src, line, sizeof(rc_data.cap_src) );

  /* Read Capture volume control, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Capture Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cap_vol, line, sizeof(rc_data.cap_vol) );

  /* Read Capture volume, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Capture Volume")) != SUCCESS )
	return( FALSE );
  rc_data.cap_lev = atoi( line );

  /* Read Playback master volume control, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Playback Master Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.mst_vol, line, sizeof(rc_data.mst_vol) );

  /* Read Playback PCM volume control, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Playback PCM Volume Control")) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.pcm_vol, line, sizeof(rc_data.pcm_vol) );

  /* Read Playback volume, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Capture Volume")) != SUCCESS )
	return( FALSE );
  rc_data.pbk_lev = atoi( line );

  /*** End of Sound Card configuration data ***/

  /* Read input tone (Rx BFO) frequency */
  if( Load_Line(line, xrittyrc, _("Input BFO Tone")) != SUCCESS )
	return( FALSE );
  rc_data.tone_freq = atoi( line );
  if( (rc_data.tone_freq < 500) ||
	  (rc_data.tone_freq > 2000) )
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading BFO Tone Freq\n"\
		  "Quit and correct xrittyrc"), QUIT );
	return( FALSE );
  }

  /* Read Baud Rate, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Baud Rate") ) != SUCCESS )
	return( FALSE );
  rc_data.baud_rate  = (double)Strtod( line );
  double temp = rc_data.dsp_rate / rc_data.baud_rate;
  rc_data.rtty_bit = (int)temp;

  /* Read AFSK frequency shift, abort if EOF */
  if( Load_Line(line, xrittyrc, _("AFSK Frequency Shift") ) != SUCCESS )
	return( FALSE );
  rc_data.afsk_shift = atoi( line );

  /* Read Reverse AFSK flag, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Reverse AFSK Enable Flag") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "reverse_afsk")), TRUE );
	SetFlag( ENABLE_USOS );
  }
  else if( strcmp(line, "no") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "reverse_afsk")), FALSE );
	ClearFlag( ENABLE_USOS );
  }
  else
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading xrittyrc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  /* Read stop bit length, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Stop Bit Length") ) != SUCCESS )
	return( FALSE );
  rc_data.stop_bit = (double)Strtod( line );

  /* Read preamble tone duration (in dsec, min 2), abort if EOF */
  if( Load_Line(line, xrittyrc, _("Preamble Tone Duration") ) != SUCCESS )
	return( FALSE );
  rc_data.tone_dur = atoi(line);
  if( rc_data.tone_dur < 2 )
	rc_data.tone_dur = 2;

  /* Tone duration now in dsp samples */
  rc_data.tone_dur = ( rc_data.tone_dur * rc_data.dsp_rate ) / 10;

  /* Read squelch threshold, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Squelch Threshold") ) != SUCCESS )
	return( FALSE );
  rc_data.sqlch_thr = atoi( line );

  /* Enter squelch level to hscale */
  hscale = Builder_Get_Object( main_window_builder, "squelch_value" );
  adjustment = gtk_range_get_adjustment( GTK_RANGE(hscale) );
  gtk_adjustment_set_value( adjustment, (gdouble)rc_data.sqlch_thr );
  rc_data.sqlch_thr *= 5;

  /* Read Enable Diversity flag, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Diversity Enable Flag") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "enable_diversity")), TRUE );
	SetFlag( ENABLE_DIVERSITY );
  }
  else if( strcmp(line, "no") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "enable_diversity")), FALSE );
	ClearFlag( ENABLE_DIVERSITY );
  }
  else
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading xrittyrc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

   /* Read USOS flag, abort if EOF */
  if( Load_Line(line, xrittyrc, _("USOS Enable Flag") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "usos")), TRUE );
	SetFlag( ENABLE_USOS );
  }
  else if( strcmp(line, "no") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "usos")), FALSE );
	ClearFlag( ENABLE_USOS );
  }
  else
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading xrittyrc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  /* Read CW encoder speed, abort if EOF */
  if( Load_Line(line, xrittyrc, _("CW Encoder Speed") ) != SUCCESS )
	return( FALSE );
  rc_data.morse_wpm = atoi( line );

  /* Read word wrap column, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Word Wrap Column") ) != SUCCESS )
	return( FALSE );
  rc_data.word_wrap = atoi( line );

  /* Read font string for text views, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Text View font string") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.font, line, sizeof(rc_data.font) );

  /*** Setup default main menu items ***/

  /* Read CAT serial port device, abort if EOF */
  if( Load_Line(line, xrittyrc, _("CAT Serial Port") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.cat_serial, line, sizeof(rc_data.cat_serial) );

  /* Read Transceiver type for  CAT, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Transceiver type for CAT") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "FT847") == 0 )
	rc_data.tcvr_type = FT847;
  else if( strcmp(line, "FT857") == 0 )
	rc_data.tcvr_type = FT857;
  else if( strcmp(line, "K2") == 0 )
	rc_data.tcvr_type = K2;
  else if( strcmp(line, "K3") == 0 )
	rc_data.tcvr_type = K3;
  else if( strcmp(line, _("NONE")) == 0 )
	rc_data.tcvr_type = NONE;
  else
  {
	rc_data.tcvr_type = NONE;
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading Transceiver type\n"\
		  "Quit and correct xrittyrc"), QUIT );
	return( FALSE );
  }

  /* Enable CAT, disable RTS */
  if( rc_data.tcvr_type == NONE )
	ClearFlag( ENABLE_CAT );
  else
  {
	SetFlag( ENABLE_CAT );
	ClearFlag( ENABLE_RTS_DTR );
  }

  /* Read RTS/DTR-Control serial port device, abort if EOF */
  if( Load_Line(line, xrittyrc, _("RTS-Control Serial Port") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rts_serial, line, sizeof(rc_data.rts_serial) );

  /* Read RTS/DTR-Control choice, abort if EOF */
  if( Load_Line(line, xrittyrc, _("RTS-Control Enable") ) != SUCCESS )
	return( FALSE );
  if( strcmp(line, "DTR") == 0 )
  {
	SetFlag( ENABLE_DTR );
	SetFlag( ENABLE_RTS_DTR );
	ClearFlag( ENABLE_CAT );
  }
  else if( strcmp(line, "RTS") == 0 )
  {
	SetFlag( ENABLE_RTS_DTR );
	ClearFlag( ENABLE_CAT );
  }
  else if( strcmp(line, "NONE") == 0 )
	ClearFlag( ENABLE_RTS_DTR );
  else
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading xrittyrc\n"\
		  "Unrecognized RTS/DTR-Control option\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  /* Read Record QSO enable flag, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Record QSOs Enable") ) != SUCCESS )
	return( FALSE );

  if( strcmp(line, "yes") == 0 )
  {
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "record_qsos")), TRUE );
	SetFlag( RECORD_QSO );
  }
  else if( strcmp(line, "no") == 0 )
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "record_qsos")), FALSE );
  else
  {
	fclose( xrittyrc );
	Error_Dialog(
		_("Error reading xrittyrc\n"\
		  "Unrecognized menu option\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  /* Read operator callsign, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Operator Callsign") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.call, line, sizeof(rc_data.call) );

  /* Read operator name, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Operator Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.name, line, sizeof(rc_data.name) );

  /* Read Zone, abort if EOF */
  if( Load_Line(line, xrittyrc, _("ITU Zone") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.zone, line, sizeof(rc_data.zone) );

  /* Read QTH, abort if EOF */
  if( Load_Line(line, xrittyrc, _("QTH Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.qth, line, sizeof(rc_data.qth) );

  /* Read QTH locator, abort if EOF */
  if( Load_Line(line, xrittyrc, _("QTH Locator") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.loc, line, sizeof(rc_data.loc) );

  /* Read Transmitter name, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Transmitter Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx, line, sizeof(rc_data.tx) );

  /* Read Transmitter power, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Transmitter Power") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx_power, line, sizeof(rc_data.tx_power) );

  /* Read Transmitter antenna, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Transmitter Antenna") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.tx_ant, line, sizeof(rc_data.tx_ant) );

  /* Read Receiver name, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Receiver Name") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx, line, sizeof(rc_data.rx) );

  /* Read Receiver N.Fig, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Receiver Noise Figure") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx_nfig, line, sizeof(rc_data.rx_nfig) );

  /* Read Receiver antenna, abort if EOF */
  if( Load_Line(line, xrittyrc, _("Receiver Antenna") ) != SUCCESS )
	return( FALSE );
  Strlcpy( rc_data.rx_ant, line, sizeof(rc_data.rx_ant) );

  /* Read propagation medium */
  if( Load_Line(line, xrittyrc, _("Propagation Medium") ) != SUCCESS )
	return( FALSE );
  Strlcpy( qso_record.via, line, sizeof(qso_record.via) );

  /* Read default remarks on QSO */
  if( Load_Line(line, xrittyrc, _("Default Remarks") ) != SUCCESS )
	return( FALSE );
  Strlcpy( qso_record.remarks, line, sizeof(qso_record.remarks) );

  /*** Read macro labels from rc file ***/
  /* Null buffer pointers just in case */
  for( idx = 0; idx < NUM_OF_LABELS; idx++ )
	macro[idx] = NULL;

  /* Preload a line */
  if( Load_Line(line, xrittyrc, _("Macro Labels") ) != SUCCESS )
	return( FALSE );

  for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ )
  {
	/* For setting macro button labels */
	GtkLabel *lbl;
	char lname[8];
	char label[LABEL_LENGTH+6];

	/* Abort if first line is not a label */
	if( line[0] != '[' )
	{
	  fprintf( stderr,
		  "xritty: First Macro line not a label\n%s\n", line );
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"First Macro line not a label\n"\
			"Quit xritty and correct"), QUIT);
	  return( FALSE );
	}

	/* Look for closing ']' */
	ln_idx = 1;
	while( (ln_idx <= LABEL_LENGTH) && (line[ln_idx] != ']') )
	  ln_idx++;

	/* Abort if label string > LABEL_LENGTH char + [] */
	if( ln_idx > LABEL_LENGTH )
	{
	  fprintf( stderr,
		  "xritty: Macro label too long\n%s\n", line );
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"Macro label too long\n"\
			"Quit xritty and correct"), QUIT);
	  return( FALSE );
	}

	/* Enter label string if all OK */
	snprintf( label, sizeof(label), "[F%d] ", lb_idx+1 );
	line[strlen(line)-1] = '\0';
	Strlcat( label, &line[1], sizeof(label) );

	/* Set macro button labels */
	snprintf( lname, sizeof(lname), "label%d", lb_idx+1 );
	lbl = GTK_LABEL( Builder_Get_Object( main_window_builder, lname) );
	gtk_label_set_text( lbl, label );

	/*** Read macros ***/
	mc_idx = tg_cnt = 0;
	do
	{
	  if( Load_Line(line, xrittyrc, _("Macros")) != SUCCESS )
		return( FALSE );

	  /* Read labels if line starts with '[' */
	  if( line[0] == '[' ) break;

	  /* Allocate memory to macro buffer */
	  ln_len = (int)strlen( line );

	  /* Ignore | end-of-macro delimeter */
	  if( line[ln_len-1] == '|' ) ln_len--;

	  if( !mem_realloc((void **)&macro[lb_idx],
			(size_t)(ln_len + mc_idx + 2)) )
		return( FALSE );

	  /* Concatenate lines to macro */
	  for( ln_idx = 0; ln_idx < ln_len; ln_idx++ )
	  {
		macro[lb_idx][mc_idx++] = line[ln_idx];

		/* Count tag delimiters */
		if( (line[ln_idx] == '<') ||
			(line[ln_idx] == '>') )
		  tg_cnt++;
	  }

	  /* Terminate Macro with LF and null */
	  macro[lb_idx][mc_idx++] = LF;
	  macro[lb_idx][mc_idx] = '\0';
	} /* do */
	while( line[ln_len] != '|'  );

	/* Abort if tag delimiters are odd number */
	if( tg_cnt & 1 )
	{
	  fprintf( stderr, "xritty: Malformed tag in Macro\n%s\n", line );
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"Malformed tag in Macro\n"\
			"Quit xritty and correct"), QUIT );
	  return( FALSE );
	}

  } /* for( lb_idx = 0; lb_idx < NUM_OF_LABELS; lb_idx++ ) */

  /*** Do the global config variables ***/
  /* Set appropriate baud rate menu item */
  {
	char btn[6];
	GtkWidget *item;

	snprintf( btn, sizeof(btn),
		"bd%d", (int)rc_data.baud_rate );
	btn[5] = '\0';
	item = Builder_Get_Object(popup_menu_builder, btn);
	if( !item )
	{
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"Invalid Baud rate\n"\
			"Quit xritty and correct"), QUIT );
	  return( FALSE );
	}
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(item), TRUE );
  }

  /* Set appropriate AFSK menu item */
  {
	char btn[8];
	GtkWidget *item;

	snprintf( btn, sizeof(btn),
		"afsk%d", (int)rc_data.afsk_shift );
	btn[7] = '\0';
	item = Builder_Get_Object(popup_menu_builder, btn);
	if( !item )
	{
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"Invalid AFSK shift\n"\
			"Quit xritty and correct"), QUIT );
	  return( FALSE );
	}
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(item), TRUE );
  }

  /* Set appropriate stop bit menu item */
  {
	char btn[8];
	GtkWidget *item;

	snprintf( btn, sizeof(btn),
		"stop%d", (int)(10.0 * rc_data.stop_bit) );
	btn[7] = '\0';
	item = Builder_Get_Object(popup_menu_builder, btn);
	if( !item )
	{
	  fclose( xrittyrc );
	  Error_Dialog(
		  _("Error reading xrittyrc\n"\
			"Invalid stop bit size\n"\
			"Quit xritty and correct"), QUIT );
	  return( FALSE );
	}
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(item), TRUE );
  }

  /* Set up CAT if enabled */
  switch( rc_data.tcvr_type )
  {
	case K2:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			Builder_Get_Object(popup_menu_builder, "k2")), TRUE );
	  break;

	case K3:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			Builder_Get_Object(popup_menu_builder, "k3")), TRUE );
	  break;

	case FT847:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			Builder_Get_Object(popup_menu_builder, "ft847")), TRUE );
	  break;

	case FT857:
	  gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
			Builder_Get_Object(popup_menu_builder, "ft857")), TRUE );
	  break;
  }

  /* Set RTS control */
  if( isFlagSet(ENABLE_RTS_DTR) )
  {
	rc_data.tcvr_type = NONE;
	ClearFlag( ENABLE_CAT );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  Builder_Get_Object(popup_menu_builder, "rts_ctl")), TRUE );
  }

  /* Set Rx and Tx text view font */
  font_desc = pango_font_description_from_string( rc_data.font );
  pango_font_description_free( font_desc );

  /* Wait for GTK to complete its tasks */
  while( g_main_context_iteration(NULL, FALSE) );

  /* Set volume spinbuttons */
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(Builder_Get_Object(
		  main_window_builder, "rxvol_spinbutton")), rc_data.cap_lev );
  gtk_spin_button_set_value( GTK_SPIN_BUTTON(Builder_Get_Object(
		  main_window_builder, "txvol_spinbutton")), rc_data.pbk_lev );

  /* Puts the tone freq in middle of waterfall */
  idx = ( 2 * ifft_data_length ) / wfall_width;
  if( idx )
	rc_data.ifft_stride = rc_data.dsp_rate / rc_data.tone_freq / idx;

  /* Initialize configuration */
  Init_Config();

  SetFlag( LTRS_SHIFT );
  Set_Shift_Label();

  fclose( xrittyrc );

  return( FALSE );
} /* End of Load_Config() */

/*------------------------------------------------------------------*/

/*  Usage()
 *
 *  Prints usage information
 */

  void
Usage( void )
{
  fprintf( stderr, "%s\n",
	  _("Usage: xritty [-hv]") );

  fprintf( stderr, "%s\n",
	  _("       -h: Print this usage information and exit.") );

  fprintf( stderr, "%s\n",
	  _("       -v: Print version number and exit.") );

} /* End of Usage() */

/*------------------------------------------------------------------------*/

/*  Cleanup()
 *
 *  Cleans up before quitting
 */

  void
Cleanup( void )
{
  /* Save outstanding records */
  if( isFlagSet(SAVE_RECORD) )
	Save_QSO_Record();

  /* Close open files */
  if( qso_record_fp != NULL )
  {
	fclose( qso_record_fp );
	qso_record_fp = NULL;
  }

  if( adif_log_fp != NULL )
  {
	fclose( adif_log_fp );
	adif_log_fp = NULL;
  }

  if( station_log_fp != NULL )
  {
	fclose( station_log_fp );
	station_log_fp = NULL;
  }

  /* Release sound card and Tcvr */
  Close_Tcvr_Serial();
  Close_Playback();
  Close_Capture();
  Close_Mixer();

} /* Cleanup( void ) */

/*------------------------------------------------------------------------*/

/*  Close_File()
 *
 *  Closes and NULL's an open file pointer
 */

  static void
Close_File( FILE **fp )
{
  if( *fp != NULL )
  {
	fclose( *fp );
	*fp = NULL;
  }
}

/*------------------------------------------------------------------------*/

/*  Open_File()
 *
 *  Opens a file
 */

  static int
Open_File( FILE **fp, char *fpath, const char *mode )
{
  /* Check if file exists */
  if( *fp != NULL ) return( FILE_OPEN );

  /* If file does not exist, it will open new */
  *fp = fopen( fpath, mode );
  if( *fp == NULL )
  {
	char mesg[MESG_SIZE];
	perror( fpath );
	snprintf( mesg, sizeof(mesg),
		_("Failed to open\n%s\n"\
		  "Quit xfhell and correct"), fpath );
	Error_Dialog( mesg, QUIT );
	return( FILE_FAIL );
  }
  else
  {
	/* Get current file position */
	long offset = ftell( *fp );
	if( offset < 0 )
	{
	  char mesg[MESG_SIZE];
	  perror( fpath );
	  snprintf( mesg, sizeof(mesg),
		  _("ftell() failed for\n%s\n"\
			"Quit xfhell and correct"), fpath );
	  Error_Dialog( mesg, QUIT );
	  return( FILE_FAIL );
	}

	/* Test if a new file was created */
	if( fseek(*fp, 0, SEEK_END) < 0 )
	{
	  char mesg[MESG_SIZE];
	  perror( fpath );
	  snprintf( mesg, sizeof(mesg),
		  _("fseek() failed for\n%s\n"\
			"Quit xfhell and correct"), fpath );
	  Error_Dialog( mesg, QUIT );
	  return( FILE_FAIL );
	}

	/* Test if a new file was created */
	if( ftell(*fp) == 0 )
	  return( FILE_NEW );
	else if( fseek(*fp, offset, SEEK_SET) < 0 ) /* Reset file position */
	{
	  char mesg[MESG_SIZE];
	  perror( fpath );
	  snprintf( mesg, sizeof(mesg),
		  _("fseek() failed for\n%s\n"\
			"Quit xfhell and correct"), fpath );
	  Error_Dialog( mesg, QUIT );
	  return( FILE_FAIL );
	}
 }

  return( FILE_OPEN );
} /* Open_File() */

/*------------------------------------------------------------------------*/

/*  File_Print()
 *
 *  Prints a string to a file
 */

  static gboolean
File_Print( FILE **fp, char *string )
{
  if( fprintf(*fp, "%s", string) != (int)strlen(string) )
  {
	perror( "xritty: fprintf()" );
	ClearFlag( SAVE_RECORD );
	Error_Dialog(
		_("Error printing to file\n"\
		  "Quit xritty and correct"), QUIT );
	Close_File( fp );
	return( FALSE );
  }

  /* Flush file */
  if( fflush(*fp) != 0 )
  {
	perror( "xritty: fflush()" );
	ClearFlag( SAVE_RECORD );
	Error_Dialog(
		_("Failed to fflush file\n"\
		  "Quit xritty and correct"), QUIT );
	return( FALSE );
  }

  return( TRUE );
} /* File_Print() */

/*------------------------------------------------------------------------*/

/*  Locate_Line()
 *
 *  Finds a line in a text file containing a given string
 */

  static gboolean
Locate_Line( char **line_idx, char *line, const char *string, FILE *fp )
{
  int err;

  /* Load lines and look for string */
  do
  {
	err = Load_Line( line, fp, _("line in Log File") );
	*line_idx = strstr(line, string);
  }
  while( (err == 0) && (*line_idx == NULL) );

  /* Usually EOF is the cause of error */
  if( err )
  {
	Error_Dialog(
		_("Error searching text file\n"\
		  "May be empty or corrupted"), QUIT );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );
} /* Locate_Line() */

/*------------------------------------------------------------------------*/

/*  Find_String()
 *
 *  Finds a string following spaces
 */

  static gboolean
Locate_String( char **line_idx, char *line, const char *string )
{
  if( (*line_idx = strstr(line, string)) == NULL )
  {
	Error_Dialog(
		_("Error searching station_log.txt\n"\
		  "May be empty or corrupted"), QUIT );
	return( FALSE );
  }

  *line_idx += strlen(string);

  return( TRUE );
} /* Find_String() */

/*------------------------------------------------------------------------*/

/*  Save_QSO_Record()
 *
 *  Saves QSO record to .adif and .txt files
 */

  gboolean
Save_QSO_Record( void )
{
  /* For opening files */
  char file_name[64];

  /* Buffer for reading and printing to files */
  char buff[REC_BUFF_SIZE];
  char *buff_idx;

  int
	num_records, /* Number of QSO records    */
	num_logged,  /* Number of QSL's printed  */
	fopen_ret;   /* Open_File() return value */

  /* Create a file for QSO ADIF log if not open */
  Strlcpy( file_name, getenv("HOME"), sizeof(file_name) );
  Strlcat( file_name, "/xritty/log.adif", sizeof(file_name) );
  fopen_ret = Open_File( &adif_log_fp, file_name, "a" );
  if( fopen_ret == FILE_NEW )
  {
	snprintf( buff, sizeof(buff),
		"QSO Log file created in ADIF v1.0 format by %s <EOH>\n\n",
		PACKAGE_STRING );
	if( !File_Print(&adif_log_fp, buff) )
	  return( FALSE );
  }
  else if( fopen_ret == FILE_FAIL )
	return( FALSE );

  /* Print QSO record to adif format file */
  snprintf( buff,  sizeof(buff),
	  "<CALL:%d>%s<QSO_DATE:8>%s<TIME_ON:4>%s\n"
	  "<FREQ:%d>%s<MODE:5>RTTY<RST_SENT:3>%3s<EOR>\n\n",
	  (int)strlen(qso_record.dx_call), qso_record.dx_call,
	  qso_record.date_adif, qso_record.time_adif,
	  (int)strlen(qso_record.freq), qso_record.freq,
	  qso_record.dx_rst );
  if( !File_Print(&adif_log_fp, buff) )
	return( FALSE );

  /* Create a file for station log if not already open */
  Strlcpy( file_name, getenv("HOME"), sizeof(file_name) );
  Strlcat( file_name, "/xritty/log.txt", sizeof(file_name) );
  fopen_ret = Open_File( &station_log_fp, file_name, "w+" );
  if( fopen_ret == FILE_NEW )
  {
	/* Print a header to log file */
	snprintf( buff,  sizeof(buff),
		_("|------------------------------------"\
		"--------------------------------------|\n"\
		"|         Station Log File - Created in"\
		" Text format by %10s          |\n"\
		"|------------------------------------"\
		"--------------------------------------|\n"),
		PACKAGE_STRING );
	if( !File_Print(&station_log_fp, buff) )
	  return( FALSE );

	/* Print an initial (0) count of records and qsl's */
	snprintf( buff,  sizeof(buff),
		_("|  Num of QSO Records:%-6d  Records Logged:%-6d"\
		"  Size of Records:%-4d  |\n"\
		"|------------------------------------"\
		"--------------------------------------|\n"),
		0, 0, RECORD_SIZE );
	if( !File_Print(&station_log_fp, buff) )
	  return( FALSE );
  }
  else if( fopen_ret == FILE_FAIL )
	return( FALSE );

  /*** Increment and print Record and QSL count ***/
  /* Go to beginning of log file */
  fseek( station_log_fp, 0, SEEK_SET );

  /* Find line with "Number of QSO Records:" */
  if( !Locate_Line(&buff_idx, buff,
		_("QSO Records:"), station_log_fp) )
	return( FALSE );
  num_records = atoi(buff_idx);

  /* Find "Number of QSO's QSL'd:" string */
  if( !Locate_String(&buff_idx, buff,
		_("Records Logged:")) )
	return( FALSE );
  num_logged = atoi(buff_idx);

  /* Go back to beginning of line and print new record */
  if( fseek(station_log_fp,
		-(long)(1 + strlen(buff)), SEEK_CUR) != 0 )
	return( FALSE );
  snprintf( buff,  sizeof(buff),
	  _("|  Num of QSO Records:%-6d  Records Logged:%-6d"\
	  "  Size of Records:%-4d  |\n"\
	  "|------------------------------------"\
	  "--------------------------------------|\n"),
	  ++num_records, num_logged, RECORD_SIZE );
  if( !File_Print(&station_log_fp, buff) )
	return( FALSE );

  /*** Print record in text log file ***/
  if( fseek(station_log_fp, 0, SEEK_END) != 0 )
	return( FALSE );
  snprintf( buff,  sizeof(buff),
	  _("|DX/ CALL: %-14s NAME: %-12s QTH: %-12s  LOC: %-6s|\n"\
	  "|MY/ CALL: %-14s ZONE: %-12s QTH: %-12s  LOC: %-6s|\n"\
	  "|QSO DATE: %-11s    TIME: %-5s UTC   FREQ: %-13s QSL: NO    |\n"\
	  "|QSO MODE: %-11s DX-REPT: %-3s      MY-REPT: %-3s     VIA: %-12s|\n"\
	  "|MY TRANS: %-11s   POWER: %-9s    ANT: %-15s          |\n"\
	  "|MY RECVR: %-11s   N.FIG: %-7s      ANT: %-15s          |\n"\
	  "| REMARKS: %-45s                   |\n"\
	  "|------------------------------------"\
	  "--------------------------------------|\n"),
	  qso_record.dx_call, qso_record.dx_name, qso_record.dx_qth,
	  qso_record.dx_loc, rc_data.call, rc_data.zone, rc_data.qth,
	  rc_data.loc, qso_record.date, qso_record.time, qso_record.freq,
	  qso_record.mode, qso_record.dx_rst, qso_record.my_rst,
	  qso_record.via, rc_data.tx, rc_data.tx_power, rc_data.tx_ant,
	  rc_data.rx, rc_data.rx_nfig, rc_data.rx_ant, qso_record.remarks );

  if( !File_Print(&station_log_fp, buff) )
	return( FALSE );

  fflush( station_log_fp );
  fflush( adif_log_fp );

  ClearFlag( SAVE_RECORD );

  return( TRUE );
} /* Save_QSO_Record() */

/*------------------------------------------------------------------------*/

/*  Process_RST()
 *
 *  Processes RST-out and RST-in fields
 */

  void
Process_RST( GtkEditable *editable )
{
  const gchar *rst;
  int len, idx;

  /* Validate RST: numerics only, */
  /* min 337 max 559, ignore blank field */
  rst = gtk_entry_get_text( GTK_ENTRY(editable) );
  len = (int)strlen(rst);

  /* Ignore blank fields */
  if( len == 0 ) return;

  for( idx = 0; idx < len; idx++ )
  {
	/* Reject non-numbers */
	if( (rst[idx] < '0') ||
		(rst[idx] > '9') )
	{
	  Bad_Entry_Dialog(
		  _("Invalid character entered\n"\
			"Numbers only allowed") );
	  return;
	}

	/* Reject RST < 337 or > 599 */
	switch( idx )
	{
	  case 0:
		if( (rst[idx] < '3') ||
			(rst[idx] > '5') )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 3 and Max = 5") );
		  return;
		}
		break;

	  case 1:
		if( rst[idx] < '3' )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 3 and Max = 9") );
		  return;
		}
		break;

	  case 2:
		if( rst[idx] < '7' )
		{
		  Bad_Entry_Dialog(
			  _("Number out of range\n"\
				"Min = 7 and Max = 9") );
		  return;
		}
	} /* switch( idx ) */

  } /* for( idx = 0; idx < len; idx++ ) */

} /* Process_RST() */

/*------------------------------------------------------------------------*/

/*  Select_Macro()
 *
 *  Selects a macro to be transmitted
 */

  void
Select_Macro( int macro_num )
{
  /* Transmit a macro */
  if( isFlagClear(TRANSMIT_MACRO) )
  {
	SetFlag( TRANSMIT_MACRO );
	ClearFlag( RECEIVE_MODE | TRANSMIT_KEYBD );
	macro_idx = macro_num - 1;
	Set_TxRx_Labels();

	/* Enable macro xmission */
	g_idle_add( Transmit_Macro, "dummy" );
  }
  else ClearFlag( TRANSMIT_MACRO );

} /* Select_Macro() */

/*------------------------------------------------------------------------*/

/*  Read_QSO_Record()
 *
 *  Reads and validates QSO record entries
 */

  gboolean
Read_QSO_Record( void )
{
  GtkWidget *entry;

  /* Enter field values to QSO record structure */
  entry = Builder_Get_Object( main_window_builder, "callsign" );
  Strlcpy( qso_record.dx_call,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_call) );
  entry = Builder_Get_Object( main_window_builder, "rst_in" );
  Strlcpy( qso_record.my_rst,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.my_rst) );
  entry = Builder_Get_Object( main_window_builder, "rst_out" );
  Strlcpy( qso_record.dx_rst,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_rst) );
  entry = Builder_Get_Object( main_window_builder, "op_name" );
  Strlcpy( qso_record.dx_name,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_name) );
  entry = Builder_Get_Object( main_window_builder, "qth_name" );
  Strlcpy( qso_record.dx_qth,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_qth) );
  entry = Builder_Get_Object( main_window_builder, "locator" );
  Strlcpy( qso_record.dx_loc,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.dx_loc) );
  entry = Builder_Get_Object( main_window_builder, "band" );
  Strlcpy( qso_record.freq,
	  gtk_entry_get_text(GTK_ENTRY(entry)),
	  sizeof(qso_record.freq) );

  /* Enter operating mode */
  Strlcpy( qso_record.mode, "RTTY", sizeof(qso_record.mode) );

  /* Validate QSO Record */
  if( (strlen(qso_record.dx_call) > 2) &&
	  (strlen(qso_record.my_rst) == 3) &&
	  (strlen(qso_record.freq)    > 0) )
	return( TRUE );
  else return( FALSE );

} /* Read_QSO_Record() */

/*------------------------------------------------------------------------*/

/***  Memory allocation/freeing utils ***/
gboolean mem_alloc( void **ptr, size_t req )
{
  free_ptr( ptr );
  *ptr = malloc( req );
  if( *ptr == NULL )
  {
	perror( "xritty: alloc():" );
	Error_Dialog( _("A memory allocation failed\n"\
		  "Please quit xritty and correct"), QUIT  );
	return( FALSE );
  }
  return( TRUE );
} /* End of void mem_alloc() */

/*------------------------------------------------------------------------*/

gboolean mem_realloc( void **ptr, size_t req )
{
  *ptr = realloc( *ptr, req );
  if( *ptr == NULL )
  {
	perror( "xritty: realloc():" );
	Error_Dialog( _("A memory re-allocation failed\n"\
		  "Please quit xritty and correct"), QUIT  );
	return( FALSE );
  }
  return( TRUE );
} /* End of void mem_realloc() */

/*------------------------------------------------------------------------*/

void free_ptr( void **ptr )
{
  if( *ptr != NULL )
	free( *ptr );
  *ptr = NULL;

} /* End of void free_ptr() */

/*------------------------------------------------------------------------*/

/* Strlcpy()
 *
 * Copies n-1 chars from src string into dest string. Unlike other
 * such library fuctions, this makes sure that the dest string is
 * null terminated by copying only n-1 chars to leave room for the
 * terminating char. n would normally be the sizeof(dest) string but
 * copying will not go beyond the terminating null of src string
 */
  void
Strlcpy( char *dest, const char *src, size_t n )
{
  char ch = src[0];
  int idx = 0;

  /* Leave room for terminating null in dest */
  n--;

  /* Copy till terminating null of src or to n-1 */
  while( (ch != '\0') && (n > 0) )
  {
	dest[idx] = src[idx];
	idx++;
	ch = src[idx];
	n--;
  }

  /* Terminate dest string */
  dest[idx] = '\0';

} /* Strlcpy() */

/*------------------------------------------------------------------*/

/* Strlcat()
 *
 * Concatenates at most n-1 chars from src string into dest string.
 * Unlike other such library fuctions, this makes sure that the dest
 * string is null terminated by copying only n-1 chars to leave room
 * for the terminating char. n would normally be the sizeof(dest)
 * string but copying will not go beyond the terminating null of src
 */
  void
Strlcat( char *dest, const char *src, size_t n )
{
  char ch = dest[0];
  int idd = 0; /* dest index */
  int ids = 0; /* src  index */

  /* Find terminating null of dest */
  while( (n > 0) && (ch != '\0') )
  {
	idd++;
	ch = dest[idd];
	n--; /* Count remaining char's in dest */
  }

  /* Copy n-1 chars to leave room for terminating null */
  n--;
  ch = src[ids];
  while( (n > 0) && (ch != '\0') )
  {
	dest[idd] = src[ids];
	ids++;
	ch = src[ids];
	idd++;
	n--;
  }

  /* Terminate dest string */
  dest[idd] = '\0';

} /* Strlcat() */

/*------------------------------------------------------------------*/

/* Strtod()
 *
 * Replaces strtod() to take into account the
 * locale-dependent decimal point character
 */
double Strtod( const char *nptr )
{
  int idx;
  size_t len;
  double d = 0.0;
  char *s = NULL;
  static gboolean first_call = TRUE;
  static char dp = '.';


  /* Find locale-dependent decimal point character */
  if( first_call )
  {
	struct lconv *lcnv;
	lcnv = localeconv();
	dp = *lcnv->decimal_point;
	first_call = FALSE;
  }

  /* Look for a . or , decimal point character
   * in the supplied number buffer (string) */
  len = strlen( nptr );
  for( idx = 0; idx < (int)len; idx++ )
	if( (nptr[idx] == ',') || (nptr[idx] == '.') )
	  break;

  /* If a decimal point character is found, replace */
  if( !mem_alloc((void **)&s, len + 1) )
	return( 0.0 );
  strncpy( s, nptr, len + 1 );
  if( idx < (int)len ) s[idx] = dp;
  d = strtod( s, NULL );
  free_ptr( (void **)&s );

  return( d );
} /* End of Strtod() */

/*------------------------------------------------------------------*/

/* Init_Config()
 *
 * Initializes runtime config variables
 */
  void
Init_Config( void )
{
  /*  Total number of signal samples of RTTY character
   *  is 1 start bit + data bits + stop bit */
  rc_data.samples_per_char =
	LEVELS_PER_BIT +
	LEVELS_PER_BIT * DATABITS_PER_CHAR +
	(int)( (double)LEVELS_PER_BIT * rc_data.stop_bit + 0.5 );

  /* Length of RTTY elements */
  double temp = (double)rc_data.dsp_rate / rc_data.baud_rate;
  rc_data.rtty_bit4    = (int)( temp / (double)LEVELS_PER_BIT );
  rc_data.rtty_bit     = rc_data.rtty_bit4 * 4;
  rc_data.rtty_2bit    = rc_data.rtty_bit * 2;
  rc_data.stop_bit_len = (int)( temp * rc_data.stop_bit );

  /* AFSK mark and space frequencies */
  double w = M_2PI / rc_data.dsp_rate;
  if( isFlagSet(REVERSE_AFSK) )
  {
	rc_data.mark_freq  = rc_data.tone_freq - rc_data.afsk_shift / 2;
	rc_data.space_freq = rc_data.tone_freq + rc_data.afsk_shift / 2;
  }
  else
  {
	rc_data.mark_freq  = rc_data.tone_freq + rc_data.afsk_shift / 2;
	rc_data.space_freq = rc_data.tone_freq - rc_data.afsk_shift / 2;
  }

  /* Omega increments */
  rc_data.mark_dW  = (double)rc_data.mark_freq  * w;
  rc_data.space_dW = (double)rc_data.space_freq * w;
  rc_data.tone_dW  = (double)rc_data.tone_freq  * w;

  Make_Cosine_Table();

} /* Init_Config() */

/*------------------------------------------------------------------*/

/* Create_Wfall_Pixbuf()
 *
 * Creates a new pixbuf for the waterfall display
 */
  gboolean
Create_Wfall_Pixbuf( int width, int height )
{
  int idx;

  /* Destroy existing pixbuff */
  if( wfall_pixbuf != NULL )
  {
	g_object_unref( G_OBJECT(wfall_pixbuf) );
	wfall_pixbuf = NULL;
  }

  /* Create waterfall pixbuf */
  wfall_pixbuf = gdk_pixbuf_new( GDK_COLORSPACE_RGB, FALSE, 8, width, height );
  if( wfall_pixbuf == NULL ) return( FALSE );

  wfall_pixels = gdk_pixbuf_get_pixels( wfall_pixbuf );
  wfall_width  = gdk_pixbuf_get_width ( wfall_pixbuf );
  wfall_height = gdk_pixbuf_get_height( wfall_pixbuf );
  wfall_rowstride  = gdk_pixbuf_get_rowstride( wfall_pixbuf );
  wfall_n_channels = gdk_pixbuf_get_n_channels( wfall_pixbuf );
  gdk_pixbuf_fill( wfall_pixbuf, 0 );

  /* Allocate average bin value buffer */
  if( !mem_realloc((void **)&bin_ave, (size_t)wfall_width * sizeof(int)) )
	return( FALSE );

  for( idx = 0; idx < wfall_width; idx++ )
	bin_ave[idx] = 0;

  /* Initialize ifft. The waterfall width has to be odd
   * to provide a center line and the FFT width has to
   * be a power of 2 so we compensate by adding 1 */
  if( ! Initialize_IFFT((int16_t)wfall_width + 1) )
	  return( FALSE );

  return( TRUE );
} /* Create_Wfall_Pixbuf() */

/*------------------------------------------------------------------*/

